require 'mime-types'
require 'webrick'
require 'uri'
require 'pp'

# Primary module
module Hyde
  # Hyde version
  # @type [String]
  VERSION = '0.5 (alpha)'
  attr_reader :VERSION

  # Hyde branding and version (for error templates)
  # @type [String]
  VLINE = "Hyde/#{Hyde::VERSION} on WEBrick/#{WEBrick::VERSION} (Ruby/#{RUBY_VERSION}/#{RUBY_RELEASE_DATE})\n"
  attr_reader :VLINE

  # Generate HTML error template
  # @param errortext [String] Error explanation
  # @param backtrace [String] Ruby backtrace
  def error_template(errortext, backtrace)
    <<~HTMLEOF
      <!DOCTYPE HTML>
      <html>
          <head>
              <title>#{WEBrick::HTMLUtils.escape(errortext)}</title>
              <style> .header {background-color: #CC7878; padding: 0.5rem; border-bottom-width: 0.2rem; border-bottom-style: solid; border-bottom-color: #202222; overflow: auto;} .title { font-weight: bolder; font-size: 36px; margin: 0 0; text-shadow: 1px 1px 1px #202222, 2px 2px 2px #404444; float: left } body { margin: 0;
      } .text { font-size 1rem; } .small { color: #7D7D7D; font-size: 12px;} .code { font-family: monospace; font-size: 0.7rem; } </style>
          </head>
          <body>
              <div class="header">
                  <p class="title">HYDE</p>
                  <p style="float: right"><a href="https://adastra7.net/git/yessiest/hyde">Source code</a></p>
              </div>
              <div style="padding: 0.5rem">
                  <p class="text">#{WEBrick::HTMLUtils.escape(errortext)}</p>
                  <pre><code class="text code">
                  #{WEBrick::HTMLUtils.escape(backtrace) or "\n\n\n"}
                  </code></pre>
                  <hr/>
                  <p class="small">#{WEBrick::HTMLUtils.escape(Hyde::VLINE)}</p>
              </div>
          </body>
      </html>
    HTMLEOF
  end
  module_function :error_template

  WEBrick::HTTPResponse.class_exec do
    public

    attr_accessor :recent_backtrace

    def create_error_page
      @body = Hyde.error_template(@reason_phrase, @recent_backtrace)
    end
  end

  # Interchangeable glob/regex/string pattern matching
  module PatternMatching
    def _prep_path(path, safe_regexp: true)
      @safe_regexp = safe_regexp
      @path = _normalize(path) if path.is_a? String
      @path = path if path.is_a? Regexp
    end

    # @return [Boolean]
    def _match?(path, _ctx)
      # behvaiour used by "index" method
      return true if @path == ''
      split_path = path.split('/').filter { |x| x != '' }
      if @path.is_a? Regexp
        # this chunk of fuck is needed to force regexp into 3 rules:
        # 1) unsafe regexp means match the whole (remaining) line.
        # 3) safe regexp means match only the part before the next slash
        # 2) a ._match? only returns when there are no leftovers
        # this forces the matching to work somewhat consistently
        test = @path.match _normalize_input(path) unless @safe_regexp
        test = @path.match split_path[0] if @safe_regexp
        if test and (test.pre_match == '') and (test.post_match == '')
          true
        else
          false
        end
      else
        # algorithm to match path segments until no more left in @path
        @path.split('/').filter { |x| x != '' }
        .each_with_index do |x, i|
          return false if x != split_path[i]
        end
        true
      end
    end

    def _normalize_input(path)
      # remove duplicate slashes and trim edge slashes
      (path.split '/').filter { |x| x != '' }.join('/')
    end

    def _normalize(path)
      # remove duplicate slashe s and trim edge slashes
      path = _normalize_input(path)
      # globbing behaviour simulated with regexp
      if path.match /(?<!\\)[\*\?\[]/
        path = Regexp.new(path
          .gsub(/[\^\$\.\|\+\(\)\{\}]/, '\\\\\\0')
          .gsub(/(?<!\\)\*\*/, '.*')
          .gsub(/(?<![.\\])\*/, '[^/]*')
          .gsub(/(?<!\\)\?/, '.')
          .gsub(/(?<!\\)\[\!/, '[^')
        )
      end
      path
    end
  end

  # Methods to control requests, accessible from within blocks
  module PublicContextControlMethods
    def redirect(url)
      res = @current_context.response
      res.status = 301
      res.body = "<HTML><A HREF=\"#{url}\">#{url}</A>.</HTML>\n"
      res.header['location'] = URI(url).to_s
      throw :controlled_exit, @current_context
    end

    def rewrite(url)
      new_context = Context.rewrite(@current_context, url)
      new_context.exit_loop = false
      throw :controlled_exit, new_context
    end

    def die(code, message = nil, backtrace = '')
      @current_context.response.status = code
      @current_context.response.backtrace = backtrace
      message ||= WEBrick::HTTPStatus::StatusMessage[code]
      if @current_context.codehandlers[code]
        @current_context.codehandlers[code].call(@current_context, message, backtrace)
      else
        @current_context.response.body = Hyde.error_template(message, backtrace)
      end
      throw :controlled_exit, @current_context
    end
  end

  # Request wrapper class
  class Context
    def initialize(path, request, response)
      @path = path
      @filepath = ''
      @request = request
      @response = response
      @indexlist = []
      @vars = {}
      @codehandlers = {}
      @queue_postprocess = []
      @queue_finalize = []
      @exit_loop = false
    end

    def self.rewrite(pctx, newpath)
      newctx = Context.new(newpath, pctx.request, pctx.response)
      newctx.vars = pctx.vars
      newctx.queue_finalize = pctx.queue_finalize.clone
      newctx.queue_postprocess = pctx.queue_postprocess.clone
      newctx
    end

    def enqueue_finalizer(dup: false, &block)
      return unless block_given?
      @queue_finalize.append(block) if dup or !@queue_finalize.include? block
    end

    def enqueue_postprocessor(&block)
      @queue_postprocess.append(block) if block_given?
    end
    attr_reader :request
    attr_reader :response
    attr_accessor :queue_finalize
    attr_accessor :queue_postprocess
    attr_accessor :path
    attr_accessor :indexlist
    attr_accessor :filepath
    attr_accessor :exit_loop
    attr_accessor :vars
    attr_accessor :codehandlers
  end

  # Context object with safe path encapsulation
  class ProtectedContext < Context
    def initialize(request)
      @path = request.path
      @filepath = request.filepath
      @request = request.request
      @response = request.response
      @indexlist = request.indexlist
      @exit_loop = request.exit_loop
      @vars = request.vars
      @codehandlers = request.codehandlers
      @queue_postprocess = request.queue_postprocess
      @queue_finalize = request.queue_finalize
    end
    undef :path=
    undef :filepath=
    undef :indexlist=
    undef :queue_postprocess=
    undef :queue_finalize=
  end

  # Handler classes
  class Probe
    include Hyde::PatternMatching
    include Hyde::PublicContextControlMethods
    def initialize(path, safe_regexp: true, &block_optional)
      _prep_path path, safe_regexp: safe_regexp
      @block = block_optional
    end

    def _match(request)
      return unless @block and (_match? request.path, request)
      @current_context = Hyde::ProtectedContext.new(request)
      return_later = instance_exec @current_context, &@block
      return_later
    end

    # @sg-ignore
    def _match?(path, request)
      # End node - nothing must be after it
      return unless super(path, request)
      match_path = _normalize_input(path).match(@path)
      match_path.post_match == ''
    end
  end

  class Serve < Hyde::Probe
    def _match(request)
      return super if @block
      return unless _match? request.path, request
      @current_context = request
      match_path = _normalize_input(request.path).match(@path)[0]
      filepath = request.filepath + match_path
      begin
        mimetype = MIME::Types.type_for(filepath)
        file = File.new filepath, 'r'
        data = file.read
        request.response.body = data
        request.response['Content-Type'] = mimetype
      rescue Errno::ENOENT
        die(404)
      end
    end
  end

  class GetMatch < Hyde::Probe
    @match_method = 'get'
    def initialize(*a, **b, &block)
      @match_method = (self.class.instance_variable_get :@match_method)
      raise Exception, 'block required!' unless block
      super(*a, **b, &block)
    end

    def _match?(path, ctx)
      if ctx.request.request_method == @match_method.upcase
        super(path, ctx)
      else
        false
      end
    end
  end

  class PostMatch < GetMatch
    @match_method = 'post'
  end

  class PutMatch < GetMatch
    @match_method = 'put'
  end

  class PatchMatch < GetMatch
    @match_method = 'patch'
  end

  class DeleteMatch < GetMatch
    @match_method = 'delete'
  end

  class OptionsMatch < GetMatch
    @match_method = 'options'
  end

  class LinkMatch < GetMatch
    @match_method = 'link'
  end

  class UnlinkMatch < GetMatch
    @match_method = 'unlink'
  end

  class PrintProbe < Hyde::Probe
    def _match(request)
      puts "#{request.path} matched!" if _match? request.path, request
    end
  end

  # Handler invocation functions
  module Handlers
    {
      probe: Hyde::Probe,
      printProbe: Hyde::PrintProbe,
      serve: Hyde::Serve,
      post: Hyde::PostMatch,
      get: Hyde::GetMatch,
      put: Hyde::PutMatch,
      patch: Hyde::PatchMatch,
      delete: Hyde::DeleteMatch,
      options: Hyde::OptionsMatch,
      link: Hyde::LinkMatch,
      unlink: Hyde::UnlinkMatch
    }.each_pair do |name, newclass|
      define_method name do |path, *a, **b, &block|
        if path.is_a? Array
          path.each do |x|
            @chain.append newclass.new x, *a, **b, &block
          end
        else
          @chain.append newclass.new path, *a, **b, &block
        end
      end
    end
  end

  class Pathspec
    include Hyde::PatternMatching
    include Hyde::Handlers
    include Hyde::PublicContextControlMethods
    def initialize(path, root_path: nil, safe_regexp: true, &block)
      _prep_path path, safe_regexp: safe_regexp
      @chain = []
      @root_override = root_path
      @remap = false
      instance_exec(&block)
    end

    def path(path, *a, **b, &block)
      if path.is_a? Array
        path.each do |x|
          @chain.append Hyde::Pathspec.new x, *a, **b, &block
        end
      else
        @chain.append Hyde::Pathspec.new path, *a, **b, &block
      end
    end

    def root(path)
      @root_override = '/' + _normalize_input(path)
    end

    def remap(path)
      @root_override = '/' + _normalize_input(path)
      @remap = true
    end

    def index(list)
      @indexlist = list if list.is_a? Array
      @indexlist = [list] if list.is_a? String
    end

    def preprocess(&block)
      @preprocessor = block
    end

    def postprocess(&block)
      @postprocessor = block
    end

    def finalize(dup: false, &block)
      @finalizer = block
      @finalizer_dup = dup
    end

    def _match(request)
      @current_context = request
      instance_exec request, &@preprocessor if @preprocessor
      request.enqueue_postprocessor(&@postprocessor) if @preprocessor
      request.enqueue_finalizer dup: @finalizer_dup, &@finalizer if @finalizer
      if _match? request.path, request
        match_path = _normalize_input(request.path).match(@path)
        next_path = match_path[0]
        request.path = cut_path = match_path.post_match
        # remap/root method handling
        if @root_override
          request.filepath = if @remap
                               @root_override + '/'
                             else @root_override + '/' + next_path + '/' end
        else
          request.filepath = request.filepath + next_path + '/'
        end
        # redefine indexing parameters if they are defined for a pathspec
        request.indexlist = @indexlist if @indexlist
        # do directory indexing
        if cut_path.match %r{^/?$}
          request.indexlist.each do |x|
            try_index = @chain.find { |y| y._match? x, request }
            if try_index
              request.path = x
              return try_index._match request
            end
          end
        end
        # passthrough to the next path object
        next_pathspec = @chain.find { |x| x._match? cut_path, request }
        next_pathspec._match request if next_pathspec
        unless next_pathspec
          # die and throw up if nowhere to go
          die(404)
        end
      end
      @current_context = nil
    end
  end
end
