module Hyde
  # String and path processing utilities
  module PatternMatching
    # Strips extra slashes from a string
    # (including slashes at the start and end of the string)
    # @param string [String]
    # @return [String]
    def self.canonicalize(string)
      string.gsub(/\/+/, "/")
            .delete_prefix("/")
            .delete_suffix("/")
    end

    # Implements glob-like pattern matching
    # Exact specifications for globbing rules:
    # "/"
    #   - act as directory separators
    #   - multiple slashes (i.e. "///") are the same as one slash ("/")
    #   - slashes are stripped at start and end of an expression or path
    #   - slashes are not matched by anything but the globstar ("**")
    #
    # "*" ( regexp: /([^/]*)/ )
    #   - matches from 0 to any number of characters
    #   - does not match nothing if placed between two slashes (i.e "/*/")
    #   - result is captured in an array
    #   - stops at slashes
    #   - greedy (matches as much as possible)
    #
    # "**" ( regexp: /(.*)/ )
    #   - matches any number of characters
    #   - matches slashes ("/")
    #   - result is captured in an array
    #   - does not stop at slashes
    #   - greedy (matches as much as possible)
    #
    # "?" ( regexp: /[^/]/ )
    #   - matches exactly one character
    #   - result is not captured
    #   - cannot match slashes
    #
    # "[...]" ( regexp: itself, ! and ^ at the start are interchangeable )
    #   - acts like a regexp range
    #   - matches any characters, including slashes if specified
    #   - valid ways to specify a range: [A-z], [a-z], [9-z] (ascii order)
    #   - ! or ^ at the start invert meaning (any character not in range)
    #   - result is not captured
    #
    # ":<name>" ( regexp: acts like a named group for /[^/]*/ )
    #   - acts like * as defined above
    #   - result is captured in a hash with <name> as key
    #   - <name> allows alphanumeric characters and underscores
    class Glob
      # @param input [String] Glob pattern
      def initialize(pattern)
        pattern = Hyde::PatternMatching.canonicalize(pattern)
        pieces = pattern.split(/(\/\*\*\/|\*\*|\*|\?|\[!?\w-\w\]|:[^\/]+)/)
        # @type [Array<String,Integer>]
        @index = build_index(pieces)
        # @type [Regexp]
        @glob = Regexp.new(pieces.map do |filter|
          case filter
          when "/**/" then "/(.*/|)"
          when "**" then "(.*)"
          when "*" then "([^/]*)"
          when "?" then "[^/]"
          when /^\[!?\w-\w\]$/ then filter.sub('!', '^')
          when /:[\w_]+/ then "[^/]*"
          else filter.gsub(/[\^$.|+(){}]/, '\\\\\\0')
          end
        end.join("").prepend("^/?"))
        puts @glob
      end

      # Match the string and assign matches to parameters
      # Returns:
      # - Unmatched part of a string
      # - Unnamed parameters
      # - Named parameters
      # @param input [String] String to match
      # @return [Array(String,Array,Hash)]
      def match(input)
        input = Hyde::PatternMatching.canonicalize(input)
        result = input.match(@glob)
        input = result.post_match
        named_params, params = assign_by_index(@index, result.captures)
        [input, params, named_params]
      end

      # Test if a string can be matched
      # Lighter version of match that doesn't assign any variables
      # @param input [String] String to match
      # @return [Boolean]
      def match?(input)
        input = Hyde::PatternMatching.canonicalize(input)
        input.match? @glob
      end

      # Test if input is convertible to a Glob and if there is any reason to
      # @param input
      # @return [Boolean] Input can be safely converted to Glob
      def self.can_convert?(input)
        input.kinf_of? String and
          input.match?(/(?<!^\\)(?:\*\*|\*|\?|\[!?\w-\w\]|:[^\/]+)/)
      end

      private

      # Build an index for separating normal matches from named matches
      # @param pieces [Array(String)] Glob pattern after being splitted
      # @return [Array(String,Integer)] Index array to use with assign_to_index
      def build_index(pieces)
        count = -1
        index = []
        pieces.each do |x|
          if x.match?(/(\*\*|\*)/)
            index.append(count += 1)
          elsif (name = x.match(/:[^\/]+/))
            index.append(name)
          end
        end
        index
      end

      # Assign values from match.captures to named and numbered groups
      # @param index [Array(String,Integer)] Index array generated by build_index
      # @param params [Array] Unnamed captures from a String.match
      def assign_by_index(index, params)
        named_params = {}
        new_params = []
        params.each_with_index do |x, k|
          if index[k].is_a? String
            named_params[index[k]] = x
          else
            new_params[index[k]] = x
          end
        end
        [named_params, new_params]
      end
    end

    # Implements regexp pattern matching
    class ReMatch
      def initialize(pattern)
        @glob = pattern
      end

      def match
        
      end

      def self.can_convert?(string)
        string.is_a? Regexp
      end
    end

    # Umbrella class that picks a suitable pattern to be a middle man for
    class Pattern
      def initialize(pattern, **options)
        @pattern = patternify(pattern)
        @static = @pattern.is_a? String
        @options = options
      end

      def static?
        @static
      end

      def match

      end

      private
      
      # Try and convert the string to a pattern, if possible
      def patternify(pattern)
        Glob.new(pattern, **@options) if Glob.can_convert?(pattern)
        ReMatch.new(pattern, **@options) if Glob.can_convert?(pattern)
      end
    end
  end

  class PathBinding
    
  end

  class Path

  end
end
