module Hyde
  module PatternMatching
    # Implements glob-like pattern matching
    # Exact specifications for globbing rules:
    # "/"
    #   - act as directory separators
    #   - multiple slashes (i.e. "///") are the same as one slash ("/")
    #   - slashes are stripped at start and end of an expression or path
    #   - slashes are not matched by anything but the globstar ("**")
    #
    # "*" ( regexp: /([^/]*)/ )
    #   - matches from 0 to any number of characters
    #   - does not match nothing if placed between two slashes (i.e "/*/")
    #   - result is captured in an array
    #   - stops at slashes
    #   - greedy (matches as much as possible)
    #
    # "**" ( regexp: /(.*)/ )
    #   - matches any number of characters
    #   - matches slashes ("/")
    #   - result is captured in an array
    #   - does not stop at slashes
    #   - greedy (matches as much as possible)
    #
    # "?" ( regexp: /[^/]/ )
    #   - matches exactly one character
    #   - result is captured
    #   - cannot match slashes
    #
    # "[...]" ( regexp: itself, ! and ^ at the start are interchangeable )
    #   - acts like a regexp range
    #   - matches any characters, including slashes if specified
    #   - valid ways to specify a range: [A-z], [a-z], [9-z] (ascii order)
    #   - ! or ^ at the start invert meaning (any character not in range)
    #   - result is captured
    #
    # ":<name>" ( regexp: acts like a named group for /[^/]*/ )
    #   - acts like * as defined above
    #   - result is captured in a hash with <name> as key
    #   - <name> allows alphanumeric characters and underscores
    class Glob
      # @param input [String] Glob pattern
      def initialize(pattern)
        pattern = Hyde::PatternMatching.canonicalize(pattern)
        pieces = pattern.split(/(\/\*\*\/|\*\*|\*|\?|\[!?\w-\w\]|:[^\/]+)/)
        # @type [Array<String,Integer>]
        @index = build_index(pieces)
        # @type [Regexp]
        @glob = Regexp.new(pieces.map do |filter|
          case filter
          when "/**/" then "/(.*/|)"
          when "**" then "(.*)"
          when "*" then "([^/]*)"
          when "?" then "([^/])"
          when /^\[!?\w-\w\]$/ then "(#{filter.sub('!', '^')})"
          when /:[\w_]+/ then "([^/]*)"
          else filter.gsub(/[\^$.|+(){}]/, '\\\\\\0')
          end
        end.join("").prepend("^/?"))
      end

      # Match the string and assign matches to parameters
      # Returns:
      # - Unmatched part of a string
      # - Unnamed parameters
      # - Named parameters
      # @param input [String] String to match
      # @return [Array(String,Array,Hash)]
      def match(input)
        input = Hyde::PatternMatching.canonicalize(input)
        result = input.match(@glob)
        input = result.post_match
        named_params, params = assign_by_index(@index, result.captures)
        [input, params, named_params]
      end

      # Test if a string can be matched
      # Lighter version of match that doesn't assign any variables
      # @param input [String] String to match
      # @return [Boolean]
      def match?(input)
        input = Hyde::PatternMatching.canonicalize(input)
        input.match? @glob
      end

      # Test if input is convertible to a Glob and if there is any reason to
      # @param input
      # @return [Boolean] Input can be safely converted to Glob
      def self.can_convert?(input)
        input.kinf_of? String and
          input.match?(/(?<!^\\)(?:\*\*|\*|\?|\[!?\w-\w\]|:[^\/]+)/)
      end

      private

      # Build an index for separating normal matches from named matches
      # @param pieces [Array(String)] Glob pattern after being splitted
      # @return [Array(String,Integer)] Index array to use with assign_to_index
      def build_index(pieces)
        count = -1
        index = []
        pieces.each do |x|
          if x.match?(/(\/\*\*\/|\*\*|\*|\?|\[!?\w-\w\])/)
            index.append(count += 1)
          elsif (name = x.match(/(?<=:)[^\/]+/))
            index.append(name[0])
          end
        end
        index
      end

      # Assign values from match.captures to named and numbered groups
      # @param index [Array(String,Integer)] Index array generated by build_index
      # @param params [Array] Unnamed captures from a String.match
      def assign_by_index(index, params)
        named_params = {}
        new_params = []
        puts index.inspect
        params.each_with_index do |x, k|
          if index[k].is_a? String
            named_params[index[k]] = x
          else
            new_params[index[k]] = x
          end
        end
        [named_params, new_params]
      end
    end
  end
end
